#if PROTOTYPE_ALLOWED

/* loadtbbp.c */
void load_data(FILE *fpcnt);
int add_block_variables(DB_ENSEMBLE_TYPE *db_ens);
int add_profile_variables(DB_ENSEMBLE_TYPE *db_ens, BBP_ENSEMBLE_TYPE *ens);
int get_prof_time_BB(YMDHMS_TIME_TYPE *ptime, BBP_ENSEMBLE_TYPE *ens);

/* t2db_bbp.c */
void fill_ancil1(ANCILLARY_1_TYPE *a, BBP_ENSEMBLE_TYPE *ens);
void fill_ancil2(ANCILLARY_2_TYPE *a, BBP_ENSEMBLE_TYPE *ens);
void fill_config1(CONFIGURATION_1_TYPE *c, BBP_CONFIGURATION_TYPE *bbp_conf);
void fill_bt(BOTTOM_TRACK_TYPE *bt, BBP_LEADER_TYPE *ld);
void fill_nav(NAVIGATION_TYPE *n, BBP_ENSEMBLE_TYPE *ens);
/* initadcp.c */
void init_ancil1(ANCILLARY_1_TYPE *a);
void init_ancil2(ANCILLARY_2_TYPE *a);
void init_access_var(ACCESS_VARIABLES_TYPE *acc);
/* tsub_bbp.c */
void set_depth(DB_ENSEMBLE_TYPE *db_ens, BBP_ENSEMBLE_TYPE *ens);
int set_block_comments(DB_ENSEMBLE_TYPE *db_ens, BBP_ENSEMBLE_TYPE *ens);
/* read_bbp.c */
int read_header(FILE *fp, BBP_HEADER_TYPE *header, long *ens_start);
int read_configuration(FILE *fp, BBP_ENSEMBLE_TYPE *ens);
int read_leader(FILE *fp, BBP_ENSEMBLE_TYPE *ens);
int read_velocity(FILE *fp, BBP_ENSEMBLE_TYPE *ens);
int read_amp(FILE *fp, BBP_ENSEMBLE_TYPE *ens);
int read_pgood(FILE *fp, BBP_ENSEMBLE_TYPE *ens);
int read_ensemble(FILE *fp, BBP_ENSEMBLE_TYPE *ens, DB_ENSEMBLE_TYPE *db_ens);
#else
/* loadtbbp.c */
void load_data();
int add_block_variables();
int add_profile_variables();
int get_prof_time_BB();

/* t2db_bbp.c */
void fill_ancil1();
void fill_ancil2();
void fill_config1();
void fill_bt();
void fill_nav();
/* initadcp.c */
void init_ancil1();
void init_ancil2();
void init_access_var();
/* tsub_bbp.c */
void set_depth();
int set_block_comments();
/* read_bbp.c */
int read_header();
int read_configuration();
int read_leader();
int read_velocity();
int read_amp();
int read_pgood();
int read_ensemble();
#endif

